"""This module handles lists"""

import itertools


class Uniques(list):
    """A list of unique items"""
    def __init__(self, items=None):
        super().__init__([])
        self.extend(items or [])

    def predicate(self, _):
        """Subclasses can exclude items which return False"""
        # pylint: disable=no-self-use
        return True

    def append(self, item):
        if item in self:
            return False
        if not self.predicate(item):
            return False
        super().append(item)
        return True

    def extend(self, items):
        result = False
        for item in items or []:
            if self.append(item):
                result = True
        return result


class UniquelyTrues(Uniques):
    """A unique list of items all being true-ish"""
    def predicate(self, item):
        return bool(item)


def de_duplicate(items):
    """Remove any duplicate item, preserving order

    >>> assert de_duplicate([1, 9, 2, 8, 1, 7, 2]) == [1, 9, 2, 8, 7]
    """
    return list(Uniques(items))


def flatten(list_of_lists):
    """Reduce a lists of lists to a list, comprehensively

    >>> assert flatten([[1, 2], [3, 4]]) == [1, 2, 3, 4]

    Thanks to Alex Martelli at
        https://stackoverflow.com/a/952952/500942
    """
    return [item for list_ in list_of_lists for item in list_]


def flatten_(list_of_lists):
    """Reduce a lists of lists to a list, functionally

    >>> assert flatten_([[1, 2], [3, 4]]) == [1, 2, 3, 4]

    Thanks to CTT at
        https://stackoverflow.com/a/716482/500942
    """
    return list(itertools.chain.from_iterable(list_of_lists))



"""Run scripts from the command_line

This module is a simplifying proxy to stdlib's sys.exit()
"""

import os
import sys
import inspect
from bdb import BdbQuit as DebugExit

import stackprinter

from pysyte.cli import arguments
from pysyte.cli import config as configuration
from pysyte.types.paths import makepath


def run(
        main_method, add_args=None, post_parse=None,
        usage=None, epilog=None, config=None):
    """Run a main_method from command line, parsing arguments

    if add_args(parser) is given it should add arguments to the parser
    if add_args(make_parser, description, epilog, usage) is given
        it can adjust help texts before making the parser
    if pre_parse() is given then call it instead of sys.argv
    if post_parse(args) is given then it is called with parsed args
        and should return them after any adjustments
    if config is true then configure that, else configure(args.prog)
        read config, as yaml, to data
        call main_method(args, data)
    else
        call main_method(args)
    """
    class Caller:
        def __init__(self):
            self.sys_args = sys.argv
            self.method = main_method
            self.add_args = add_args
            self.module = inspect.getmodule(main_method)
            self.doc = self.module.__doc__
            self.in_main_module = self.module.__name__ == '__main__'
            self.argcount = bool(self.method.__code__.co_argcount)
            self.needs_one_arg = self.argcount == 1
            self.needs_args = bool(self.argcount)
            self.adds_args = bool(self.add_args)

        def arg_parser(self):
            try:
                return self.add_args(
                    arguments.parser, self.doc, usage, epilog)
            except TypeError:
                try:
                    return self.add_args(
                        arguments.parser(self.doc, usage, epilog))
                except TypeError:
                    raise NotImplementedError(
                        'Unknown signature for add_args()')

        @property
        def config_name(self):
            if isinstance(config, str):
                return config
            name = makepath(self.method).name
            return name

        def main(self, argument_handler):
            if config:
                if self.needs_args:
                    argument_handler.add_args()
                config_ = configuration.all(self.config_name)
                return caller.method(self.args, config_)
            if self.needs_args:
                argument_handler.add_args()
                return self.method(self.args)
            return self.method()

    caller = Caller()

    class ArgumentHandler:

        if caller.adds_args:
            def add_args(self):
                parser = caller.arg_parser()
                assert parser  # Dont trust callers' coder to return the parser
                global args  # leave parsed args available from this module
                caller.args = args = parser.parse_args(post_parser=post_parse)
        else:
            def add_args(self):
                assert caller.needs_one_arg
                caller.args = sys.argv[1:]

        def run(self):
            """Handle expected exceptions"""
            try:
                return caller.main(self)
            except KeyboardInterrupt:
                sys.stderr.write('^c ^C ^c    ^C ^c ^C    ^c ^C ^c\n')
                ctrl_c = 3
                return ctrl_c
            except DebugExit:
                return os.X_OK
            except SystemExit as e:
                return e.code
            except Exception as e:  # pylint: disable=broad-except
                stackprinter.show(e, style='darkbg')

    if caller.in_main_module:
        handler = ArgumentHandler()
        sys.exit(handler.run())


args = {}



"""Handle punctuation for pysyte"""

space = ' '

bang = '!'
double_quote = '"'
hash_ = '#'
dollar = '$'
modulo = '%'
ampersand = '&'
apostrophe = single_quote = "'"
open_parens = '('
close_parens = ')'
asterisk = '*'
plus = '+'
comma = ','
minus = '-'
dot = full_stop = period = '.'
forward_slash = '/'
zero = '0'
colon = ':'
semi_colon = ';'
less = '<'
equal = '='
greater = '>'
question = '?'
at = '@'
open_bracket = '['
backslash = '\\'
close_bracket = ']'
caret = '^'
underline = '_'
backtick = '`'
open_brace = '{'
pipe = '|'
close_brace = '}'

punctuation = f'!"#$%&{apostrophe})(*+,-./:;<=>?@[\\]^_`{open_brace}|{close_brace}'




class Server(object):
    pass


class Protocol(object):
    def __init__(self, server):
        pass


class File(Protocol):
    name = 'file'


class Http(Protocol):
    name = 'http'


class Https(Protocol):
    name = 'https'


class Git(Protocol):
    name = 'git'


class Ssh(Protocol):
    name = 'ssh'


class Tcp(Protocol):
    name = 'tcp'



"""Classes for configuration"""

import yaml

from pysyte.types import paths
from pysyte.types.dictionaries import NameSpaces


class YamlConfiguration(NameSpaces):
    """Read a yaml config file and parse it to attributes"""
    def __init__(self, string):
        if not string:
            super().__init__({})
        else:
            path = self.as_path(string)
            with open(path) as stream:
                data = yaml.safe_load(stream)
                super(YamlConfiguration, self).__init__(data)

    def as_path(self, string):
        stem = paths.path(string)
        extensions = ('yml', 'yaml')
        for extension in extensions:
            yml = stem.add_missing_ext(extension)
            if yml.isfile():
                return yml


class YamlyConfiguration(YamlConfiguration):
    pass


class ModuleConfiguration(YamlConfiguration):
    def as_path(self, string):
        path = paths.path(string)
        return str(path.extend_by('yaml'))


class IniConfiguration(object):
    pass



"""Handle configs from program name"""

from pysyte.unix import root

from pysyte.config import xdg

def user(name):
    return xdg.home_config(name)


def machine(name):
    return xdg.etc_config(name)


def all(name):
    return {**machine(name), **user(name)}



from pysyte.cli.arguments import ArgumentsParser
from pysyte.types.paths import path


class PathParser(ArgumentsParser):
    """Handle positional args as paths"""

    def post_parser(self, args):
        for arg in args:
            arg_path = path(arg)
            if not arg_path.exists():
                continue
            args.set(arg, arg_path)
        return args



"""A collection of extended path classes and methods

The classes all inherit from the original path.path
"""
import os
import re
import stat
import importlib
from fnmatch import fnmatch
from functools import singledispatch
from functools import total_ordering

from pysyte.types.lists import flatten


class PathError(Exception):
    """Something went wrong with a path"""
    prefix = 'Path Error'

    def __init__(self, message):
        Exception.__init__(self, message)


class PathAssertions:
    """Assertions that can be made about paths"""
    # pylint: disable=no-member

    def assert_exists(self):
        """Raise a PathError if this path does not exist on disk"""
        if not self.exists():
            raise PathError(f'{self} does not exist')
        return self

    def assert_isdir(self):
        """Raise a PathError if this path is not a directory on disk"""
        if not self.isdir():
            raise PathError(f'{self} is not a directory')
        return self

    def assert_isfile(self):
        """Raise a PathError if this path is not a file on disk"""
        if not self.isfile():
            raise PathError(f'{self} is not a file')
        return self


from path import Path as PPath  # pylint: disable=wrong-import-order,wrong-import-position


@total_ordering
class NonePath:
    def __init__(self, string=None):
        self.string = string if string else ''
        self.proxy = DirectPath(self.string)

    def __str__(self):
        return self.string

    def __repr__(self):
        return f'<{self.__class__.__name__} "{str(self)}">'

    def __bool__(self):
        return False

    def __eq__(self, other):
        if self.string:
            return str(self) == str(other)
        return not other

    def __lt__(self, other):
        if self.string and other:
            return str(self) < str(other)
        return bool(other)

    def __contains__(self, _):
        return False

    def __div__(self, child):
        result = os.path.join(self.string, child) if child else self.string
        return makepath(result)

    __truediv__ = __div__

    @property
    def parent(self):
        if '/' not in self.string:
            return None
        parent_string = '/'.join(self.string.split('/')[:-1])
        return makepath(parent_string)

    def exists(self):
        return False

    isdir = isfile = exists

    def __getattr__(self, name):
        return getattr(self.proxy, name, None)


@total_ordering
class DotPath(PPath):
    """Some additions to the classic path class"""
    # pylint: disable=abstract-method
    # pylint: disable=too-many-public-methods

    def __hash__(self):
        return hash(str(self))

    def __repr__(self):
        string = repr(f'{self}')
        return f'<{self.__class__.__name__} {string}>'

    # The / operator joins paths.
    def __div__(self, child):
        """Join two path components, adding a separator character if needed.

        If the result is a file return self.__file_class__(result)

        >>> p = DotPath('/home/guido')
        >>> p.__div__('fred') == p / 'fred' == p.joinpath('fred')
        True
        """
        string = str(self)
        result = os.path.join(string, child) if child else string
        return makepath(result)

    __truediv__ = __div__

    def __eq__(self, other):
        return str(self) == str(other)

    def __lt__(self, other):
        return str(self) < str(other)

    def basename(self):
        return str(super().basename())

    @property
    def name(self):
        return str(super().name)

    def parent_directory(self):
        if self.isroot():
            return None
        return self.parent

    def parent_directories(self):
        if self.isroot():
            return []
        parent = self.parent
        return [parent] + parent.parent_directories()

    def directory(self):
        """Return a path to the path's directory"""
        return self.parent

    def dirnames(self):
        """Split the dirname into individual directory names

        An absolute path starts with an empty string, a relative path does not

        >>> DotPath('/path/to/module.py').dirnames() == ['/', 'path', 'to']
        True
        >>> DotPath('path/to/module.py').dirnames() == ['path', 'to']
        True
        """
        return [str(_) for _ in self.directory().split(os.path.sep)]

    def dirpaths(self):
        """Split the dirname into individual directory names

        An absolute path starts with an empty string, a relative path does not

        >>> p = DotPath('/path/to/x.py')
        >>> assert p.paths == p.dirpaths()
        """
        parts = self.split()
        result = [DotPath(parts[0] or '/')]
        for name in parts[1:]:
            result.append(result[-1] / name)
        return result

    def directories(self):
        """Split the dirname into individual directory names

        No empty parts are included

        >>> DotPath('path/to/module.py').directories() == ['path', 'to']
        True
        >>> DotPath('/path/to/module.py').directories() == ['/', 'path', 'to']
        True
        """
        return [d for d in self.dirnames() if d]

    parents = property(
        dirnames, None, None,
        """ This path's parent directories, as a list of strings.

        >>> DotPath('/path/to/module.py').parents == ['/', 'path', 'to']
        True
        """)

    paths = property(
        dirpaths, None, None,
        """ This path's parent directories, as a sequence of paths.

        >>> paths = DotPath('/usr/bin/vim').paths
        >>> paths[-1].isfile()  # vim might be a link
        True
        >>> paths[-2] == paths[-1].parent
        True
        >>> paths[-3] == paths[-2].parent
        True
        >>> paths[-4] == paths[-3].parent
        True
        >>> paths[-4] == paths[0]
        True
        """)

    def split(self, sep=None, maxsplit=-1):
        separator = sep or os.path.sep
        parts = super().split(separator, maxsplit)
        parts[0] = makepath(parts[0] if parts[0] else '/')
        return parts

    def abspath(self):
        return makepath(os.path.abspath(str(self)))

    def short_relative_path_to(self, destination):
        """The shorter of either the absolute path of the destination,
            or the relative path to it

        >>> print(DotPath('/home/guido/bin').short_relative_path_to(
        ...     '/home/guido/build/python.tar'))
        ../build/python.tar
        >>> print(DotPath('/home/guido/bin').short_relative_path_to(
        ...     '/mnt/guido/build/python.tar'))
        /mnt/guido/build/python.tar
        """
        relative = self.relpathto(destination)
        absolute = self.__class__(destination).abspath()
        if len(str(relative)) < len(str(absolute)):
            return relative
        return absolute

    def short_relative_path_to_here(self):
        """A short path relative to current working directory"""
        return self.short_relative_path_to(os.getcwd())

    def short_relative_path_from_here(self):
        """A short path relative to self to the current working directory"""
        return self.__class__(os.getcwd()).short_relative_path_to(self)

    def fnmatch_basename(self, glob):
        if glob.startswith(os.path.sep):
            glob = glob.lstrip(os.path.sep)
        string = self.basename()
        if fnmatch(string, glob):
            return self
        return None

    def fnmatch_directory(self, glob):
        if glob.startswith(os.path.sep) or glob.endswith(os.path.sep):
            glob = glob.strip(os.path.sep)
        if self.isdir():
            string = self.basename()
        else:
            string = self.parent.basename()
        if fnmatch(string, glob):
            return self
        return None

    def fnmatch_directories(self, glob):
        if glob.startswith(os.path.sep) or glob.endswith(os.path.sep):
            glob = glob.strip(os.path.sep)
        strings = reversed(self.directory().splitall()[1:])
        for string in strings:
            if fnmatch(string, glob):
                return self
        return None

    def fnmatch_part(self, glob):
        if self.fnmatch(glob):
            return self
        if self.fnmatch_basename(glob):
            return self
        if self.fnmatch_directory(glob):
            return self
        if self.fnmatch_directories(glob):
            return self
        return None

    def __get_owner_not_implemented(self):
        """Remove pylint warning"""
        pass

    def expand(self):
        _expand = lambda x : os.path.expanduser(os.path.expandvars(x))
        try:
            p = makepath(_expand(self))
            return p.realpath().abspath()
        except AttributeError:
            assert False, str(self.__class__(
                os.path.abspath(os.path.realpath(
                    os.path.expanduser(os.path.expandvars(str(
                        self
            )))))))

    def same_path(self, other):
        return self.expand() == other.expand()

    def isroot(self):
        raise NotImplementedError('Only a directory path can be a root')

    def ishidden(self):
        """A 'hidden file' has a name starting with a '.'"""
        s = str(self.basename())
        return s and s[0] == '.'

    def is_executable(self):
        """Whether the path is executable"""
        # pylint: disable=no-self-use
        return False

    def isexec(self):
        return self.is_executable()

    def add_ext(self, *args):
        """Join all args as extensions

        Strip any leading `.` from args

        >>> source = makepath(__file__)
        >>> new = source.add_ext('txt', '.new')
        >>> assert new.name.endswith('.py.txt.new')
        """
        exts = [(a[1:] if a[0] == '.' else a) for a in args]
        string = '.'.join([self] + list(exts))
        return makepath(string)

    def add_missing_ext(self, ext):
        """Add that extension, if it is missing

        >>> fred = makepath('fred.py')
        >>> assert fred.add_missing_ext('.py') == fred
        >>> assert fred.add_missing_ext('.txt').endswith('.py.txt')
        """
        copy = self[:]
        _stem, old = os.path.splitext(copy)
        extension = f'.{ext.lstrip(".")}'
        if old == extension:
            return makepath(self)
        return self.add_ext(extension)

    def has_executable(self):
        """Whether the path has any executable bits set """
        executable_bits = stat.S_IEXEC | stat.S_IXGRP | stat.S_IXOTH
        try:
            return bool(os.stat(self).st_mode & executable_bits)
        except OSError:
            return False


def ext_language(ext, exts=None, simple=True):
    """Language of the extension in those extensions

    If exts is supplied, then restrict recognition to those exts only
    If exts is not supplied, then use all known extensions

    >>> ext_language('.py') == 'python'
    True
    """
    languages = {
        '.py': 'python',
        '.py2': 'python' if simple else 'python2',
        '.py3': 'python' if simple else 'python3',
        '.sh': 'bash',
        '.bash': 'bash',
        '.pl': 'perl',
        '.txt': 'english',
    }
    ext_languages = {_: languages[_] for _ in exts} if exts else languages
    return ext_languages.get(ext)


del PPath


class FilePath(DotPath, PathAssertions):
    """A path to a known file"""

    def __div__(self, child):
        raise PathError('%r has no children' % self)

    __truediv__ = __div__

    def __iter__(self):
        for line in self.stripped_lines():
            yield line

    def try_remove(self):
        """Try to remove the file"""
        self.remove()

    def stripped_lines(self):
        """A list of all lines without trailing whitespace

        If lines can not be read (e.g. no such file) then an empty list
        """
        try:
            return [l.rstrip() for l in self.lines(retain=False)]
        except (IOError, UnicodeDecodeError):
            return []

    def stripped_whole_lines(self):
        """A list of all lines without trailing whitespace or blank lines"""
        return [l for l in self.stripped_lines() if l]

    def non_comment_lines(self):
        """A list of all non-empty, non-comment lines"""
        return [l
                for l in self.stripped_whole_lines()
                if not l.startswith('#')]

    def isroot(self):
        """A file cannot be root of a filesystem"""
        return False

    def is_executable(self):
        """Whether the file has any executable bits set"""
        return self.has_executable()

    def has_line(self, string):
        for line in self:
            if string == line:
                return True
        return False

    def any_line_has(self, string):
        for line in self:
            if string in line:
                return True
        return False

    def split_exts(self):
        """Split all extensions from the path

        >>> p = FilePath('here/fred.tar.gz')
        >>> assert p.split_exts() == ('here/fred', '.tar.gz')
        """
        copy = self[:]
        filename, ext = os.path.splitext(copy)
        if ext == '.gz':
            filename, ext = os.path.splitext(filename)
            ext = f'{ext}.gz'
        return self.__class__(filename), ext

    split_all_ext = split_ext = split_exts

    def as_python(self):
        """The path to the file with a .py extension

        >>> FilePath('/path/to/fred.txt').as_python()
        <FilePath '/path/to/fred.py'>
        """
        return self.extend_by('.py')

    def extend_by(self, extension):
        """The path to the file changed to use the given extension

        >>> FilePath('/path/to/fred').extend_by('.txt')
        <FilePath '/path/to/fred.txt'>
        >>> FilePath('/path/to/fred.txt').extend_by('..tmp')
        <FilePath '/path/to/fred.tmp'>
        >>> FilePath('/path/to/fred.txt').extend_by('fred')
        <FilePath '/path/to/fred.fred'>
        """
        copy = self[:]
        filename, _ = os.path.splitext(copy)
        return self.__class__(f'{filename}.{extension.lstrip(".")}')


    def make_read_only(self):
        """chmod the file permissions to -r--r--r--"""
        self.chmod(ChmodValues.readonly_file)

    def cd(self):  # pylint: disable=invalid-name
        """Change program's current directory to self"""
        return cd(self.parent)

    def dirname(self):
        return DirectPath(os.path.dirname(self))
    parent = property(dirname)

    def shebang(self):
        """The  #! entry from the first line of the file

        If no shebang is present, return an empty string
        """
        try:
            first_line = self.lines(retain=False)[0]
        except (OSError, UnicodeDecodeError):
            return ''
        if first_line.startswith('#!'):
            return first_line[2:].strip()
        return ''

    def mv(self, destination):  # pylint: disable=invalid-name
        return self.move(destination)

    def make_file_exist(self):
        """Make sure the parent directory exists, then touch the file"""
        self.parent.make_directory_exist()
        self.parent.touch_file(self.name)
        return self

    @property
    def language(self):
        """The language of this file"""
        try:
            return self._language
        except AttributeError:
            self._language = ext_language(self.ext)
        return self._language

    @language.setter
    def language(self, value):
        self._language = value


class DirectPath(DotPath, PathAssertions):
    """A path which knows it might be a directory

    And that files are in directories
    """

    __file_class__ = FilePath

    def __iter__(self):
        for a_path in DotPath.listdir(self):
            yield a_path

    def __contains__(self, thing):
        return self.contains(thing)

    def contains(self, thing):
        try:
            _ = thing.contains
            return str(thing).startswith(str(self))
        except AttributeError:
            return thing in str(self)

    def directory(self):
        """Return a path to a directory.

        Either the path itself (if it is a directory), or its parent)
        """
        if self.isdir():
            return self
        return self.parent

    def try_remove(self):
        """Try to remove the path

        If it is a directory, try recursive removal of contents too
        """
        if self.islink():
            self.unlink()
        elif self.isfile():
            self.remove()
        elif self.isdir():
            self.empty_directory()
            if self.isdir():
                self.rmdir()
        else:
            return False
        return True

    def empty_directory(self):
        """Remove all contents of a directory

        Including any sub-directories and their contents"""
        for child in self.walkfiles():
            child.remove()
        for child in reversed([self.walkdirs()]):
            if child == self or not child.isdir():
                continue
            child.rmdir()

    def cd(self):  # pylint: disable=invalid-name
        """Change program's current directory to self"""
        return cd(self)

    def listdir(self, pattern=None):
        return [self._next_class(_)
                for _ in DotPath.listdir(self, pattern)]

    def list_dirs(self, pattern=None):
        return self.list_dirs_files(pattern)[0]

    def list_files(self, pattern=None):
        return self.list_dirs_files(pattern)[1]

    def list_dirsfiles(self, pattern=None):
        dirs, others = self.list_dirs_files(pattern)
        return dirs + others

    def list_dirs_files(self, pattern=None):
        items = self.listdir(pattern)
        dirs = [_ for _ in items if _.isdir()]
        others = [_ for _ in items if not _.isdir()]
        return dirs, others

    def make_directory_exist(self):
        if self.isdir():
            return False
        if os.path.exists(self):
            raise PathError(f'{self} exists but is not a directory')
        self.makedirs()
        return True

    def make_file_exist(self, filename=None):
        """Make the directory exist, then touch the file

        If the filename is None, then use self.name as filename
        """
        if filename is None:
            path_to_file = FilePath(self)
            path_to_file.make_file_exist()
            return path_to_file
        self.make_directory_exist()
        path_to_file = self.touch_file(filename)
        return FilePath(path_to_file)

    def make_read_only(self):
        """chmod the directory permissions to -r-xr-xr-x"""
        self.chmod(ChmodValues.readonly_directory)

    def touch_file(self, filename):
        """Touch a file in the directory"""
        path_to_file = self.__file_class__(os.path.join(self, filename))
        path_to_file.touch()
        return path_to_file

    def existing_sub_paths(self, sub_paths):
        """Those in the given list of sub_paths which do exist"""
        paths_to_subs = [self / _ for _ in sub_paths]
        return [_ for _ in paths_to_subs if _.exists()]

    def clear_directory(self):
        """Make sure the directory exists and is empty"""
        self.make_directory_exist()
        self.empty_directory()

    # pylint: disable=arguments-differ
    def walkdirs(self, pattern=None, errors='strict', ignores=None):
        ignored = ignore_fnmatches(ignores)
        for path_to_dir in super(DirectPath, self).walkdirs(pattern, errors):
            if not ignored(path_to_dir.relpath(self)):
                yield path_to_dir

    # pylint: disable=arguments-differ
    def walkfiles(self, pattern=None, errors='strict', ignores=None):
        ignored = ignore_fnmatches(ignores)
        for path_to_file in super(DirectPath, self).walkfiles(pattern, errors):
            if not ignored(path_to_file.relpath(self)):
                yield path_to_file

    def listfiles(self, pattern=None, ignores=None):
        ignored = ignore_fnmatches(ignores)
        return [_ for _ in self.listdir(pattern) if _.isfile() and not ignored(_)]

    def has_vcs_dir(self):
        for vcs_dir in ('.git', '.svn', '.hg'):
            if self.fnmatch_part(vcs_dir):
                return True
        return False

    def isroot(self):
        return str(self) == '/'


def ignore_fnmatches(ignores):

    def ignored(a_path):
        if not ignores:
            return False
        for ignore in ignores:
            if a_path.fnmatch_part(ignore):
                return True
        return False

    return ignored


class ChmodValues:
    # pylint: disable=too-few-public-methods
    readonly_file = 0o444
    readonly_directory = 0o555


def _make_module_path(arg):
    """Make a path from a thing that has a module

    classes and functions have modules, they'll be needing this
    """
    try:
        return makepath(importlib.import_module(arg.__module__))
    except (AttributeError, ModuleNotFoundError):
        return None

@singledispatch
def makepath(arg):
    attribute = getattr(arg, 'path', False)
    return makepath(attribute) if attribute else makepath(str(arg))


path = makepath


@makepath.register(type(None))
def _(arg):
    return NonePath()


@makepath.register(DotPath)
def _(arg):
    return arg


@makepath.register(str)
def _(arg):
    """Make a path from a string

    Expand out any variables, home squiggles, and normalise it
    See also http://stackoverflow.com/questions/26403972
    """
    if not arg:
        return NonePath()
    if os.path.isfile(arg):
        return FilePath(arg)
    if os.path.isdir(arg):
        string = arg if arg == '/' else arg.rstrip('/')
        return DirectPath(string)
    v = os.path.expandvars(arg)
    u = os.path.expanduser(v)
    if arg == u:
        return NonePath(arg)
    if os.path.exists(u):
        return makepath(u)
    return NonePath(arg)


@makepath.register(type(os))
def _(arg):
    """Make a path from a module"""
    if arg == 'builtins':
        return None
    return makepath(arg.__file__)


@makepath.register(type(makepath))
def _(arg):
    """Make a path from a function's module"""
    terminal_regexp = re.compile('<(stdin|.*python-input.*)>')
    method = getattr(arg, '__wrapped__', arg)
    filename = method.__code__.co_filename
    if terminal_regexp.match(filename):
        return None
    return _make_module_path(method)


@makepath.register(type(DotPath))
def _(arg):
    """Make a path from a class's module"""
    return _make_module_path(arg)


def makestr(string: str):
    """Make a path from a string"""
    if os.path.isfile(string) or os.path.isdir(string):
        return makepath(string)
    return NonePath(string)


def cd(path_to):  # pylint: disable=invalid-name
    """cd to the given path

    If the path is a file, then cd to its parent directory

    Remember current directory before the cd
        so that we can cd back there with cd('-')
    """
    if path_to == '-':
        if not cd.previous:
            raise PathError('No previous directory to return to')
        return cd(cd.previous)
    if not hasattr(path_to, 'cd'):
        path_to = makepath(path_to)
    try:
        previous = os.getcwd()
    except OSError as e:
        if 'No such file or directory' in str(e):
            return False
        raise
    if path_to.isdir():
        os.chdir(path_to)
    elif path_to.isfile():
        os.chdir(path_to.parent)
    elif not path_to.exists():
        return False
    else:
        raise PathError(f'Cannot cd to {path_to}')
    cd.previous = previous
    return True


try:
    cd.previous = makepath(os.getcwd())
except (OSError, AttributeError):
    cd.previous = None


def as_path(string_or_path):
    """Return the argument as a DirectPath

    If it is already one, return it unchanged
    If not, return the makepath()
    """
    if isinstance(string_or_path, DirectPath):
        return string_or_path
    return makepath(string_or_path)


def string_to_paths(string):
    for c in ':, ;':
        if c in string:
            return strings_to_paths(string.split(c))
    return [makepath(string)]


def strings_to_paths(strings):
    return [makepath(s) for s in strings]


def paths(strings):
    return [p for p in strings_to_paths(strings) if p.exists()]


def split_directories(strings):
    paths_ = strings_to_paths(strings)
    return [_
            for _ in paths_
            if _.isdir()], [_ for _ in paths_ if not _.isdir()]


def split_files(strings):
    paths_ = strings_to_paths(strings)
    return ([_ for _ in paths_ if _.isfile()],
            [_ for _ in paths_ if not _.isfile()])


def split_directories_files(strings):
    paths_ = strings_to_paths(strings)
    return ([_ for _ in paths_ if _.isdir()],
            [_ for _ in paths_ if _.isfile()],
            [_ for _ in paths_ if not (_.isfile() or _.isdir())])


def files(strings):
    return split_files(strings)[0]


def directories(strings):
    return split_directories(strings)[0]


def root():
    return makepath('/')


def tmp():
    return makepath('/tmp')


def home():
    _home = makepath(os.path.expanduser('~'))
    assert _home
    _x = _home.expand()
    return _home


def pwd():
    return makepath(os.getcwd())


def first_dir(path_string):
    """Get the first directory in that path

    >>> first_dir('usr/local/bin') == 'usr'
    True
    """
    parts = path_string.split(os.path.sep)
    return parts[0]


def first_dirs(path_strings):
    """Get the roots of those paths

    >>> first_dirs(['usr/bin', 'bin']) == ['usr', 'bin']
    True
    """
    return [first_dir(_) for _ in path_strings]


def unique_first_dirs(path_strings):
    """Get the unique roots of those paths

    >>> unique_first_dirs(['usr/local/bin', 'bin']) == set(['usr', 'bin'])
    True
    """
    return set(first_dirs(path_strings))


def _names_in_directory(path_to_directory):
    """Get all items in the given directory

    Swallow errors to give an empty list
    """
    try:
        return os.listdir(path_to_directory)
    except OSError:
        return []


def fnmatcher(pattern, path_to_directory=None, wanted=lambda x: True):
    """Gives a method to check if an item matches the pattern, and is wanted

    If path_to_directory is given then items are checked there
    By default, every item is wanted
    """
    return lambda x: fnmatch(x, pattern) and wanted(
        os.path.join(path_to_directory, x) if path_to_directory else x)


def list_items(path_to_directory, pattern, wanted=lambda x: True):
    """All items in the given path which match the given glob and are wanted

    By default, every path is wanted
    """
    path_ = makepath(path_to_directory)
    if not path_:
        return set()
    needed = fnmatcher(pattern, path_, wanted)
    return [os.path.join(path_, name)
            for name in _names_in_directory(path_)
            if needed(name)]


def list_sub_directories(path_to_directory, pattern):
    """All sub-directories of the given directory matching the given glob"""
    return list_items(path_to_directory, pattern, os.path.isdir)


def set_items(path_to_directory, pattern, wanted):
    return set(list_items(path_to_directory, pattern, wanted))


def set_files(path_to_directory, pattern):
    """A list of all files in the given directory matching the given glob"""
    return set_items(path_to_directory, pattern, os.path.isfile)


def contains_glob(path_to_directory, pattern, wanted=lambda x: True):
    """Whether the given path contains an item matching the given glob"""
    return any(list_items(path_to_directory, pattern, wanted))


def contains_directory(path_to_directory, pattern):
    """Whether the given path contains a directory matching the given glob"""
    return contains_glob(path_to_directory, pattern, os.path.isdir)


def contains_file(path_to_directory, pattern):
    """Whether the given directory contains a file matching the given glob"""
    return contains_glob(path_to_directory, pattern, os.path.isfile)


def environ_paths(key):
    return [makepath(_) for _ in os.environ[key].split(':')]


def environ_path(key):
    return makepath(os.environ[key])


def default_environ_path(key, default):
    return makepath(os.environ.get(key, default))


def add_star(string):
    """Add '.*' to string

    >>> assert add_star('fred') == 'fred.*'
    """
    suffix = '*' if '.' in string else '.*'
    return f'{string}{suffix}'

def add_stars(strings):
    """Add '.*' to each string

    >>> assert add_stars(['fred', 'Fred.']) == ['fred.*', 'Fred.*']
    """
    paths_ = [strings] if isinstance(strings, str) else strings
    return [add_star(p) for p in paths_]


def tab_complete(strings, globber=add_stars, select=os.path.exists):
    """Finish path names "left short" by bash's tab-completion

    strings is a string or strings
        if any string exists as a path return those as paths

    globber is a method which should return a list of globs
        default: add_stars(['fred.']) == ['fred.*']
            or, e.g: add_python(['fred', 'fred.']) == ['fred.py*']
        if any expanded paths exist return those

    select is a method to choose wanted paths
        Defaults to selecting existing paths
    """
    strings_ = [strings] if isinstance(strings, str) else strings
    globs = flatten([globber(s) for s in strings_])
    here_ = pwd()
    matches = []
    for glob_ in globs:
        if '/' in glob_:
            directory, base = os.path.split(glob_)
            dir_ = here_ / directory
        else:
            dir_ = here_
            base = glob_
        match = [p for p in dir_.listdir() if p.fnmatch_basename(base)]
        matches.extend(match)
    result = [p for p in set(matches) if select(p)]
    return result if result[1:] else strings


def pyc_to_py(path_to_file):
    """Change some file extensions to those which are more likely to be text

    >>> pyc_to_py('vim.pyc') == 'vim.py'
    True
    """
    stem, ext = os.path.splitext(path_to_file)
    if ext == '.pyc':
        return f'{stem}.py'
    return path_to_file



"""Handle arguments from the command line

This module is a simplifying proxy to stdlib's argparse
"""

import argparse
import re
import shlex
import sys
from functools import partial
from itertools import chain

from deprecated import deprecated

from pysyte.cli.config import user
from pysyte.types.numbers import inty


def config(arguments):
    return user(arguments.prog)

def extract_strings(names, name):
    result = {}
    try:
        value = names[name]
    except KeyError:
        return []
    if isinstance(value, str):
        return [value]
    try:
        return [v for v in value if isinstance(v, str)]
    except TypeError:
        return []

class IntyAction(argparse.Action):
    def __call__(self, parser, namespace, value, option_string=None):
        setattr(namespace, self.dest, inty(value))


class ArgumentsParser(object):
    """Add more attriubtes and methods to an argparse.ArgumentParser"""

    def __init__(self, argparser):
        self.parser = argparser
        self._parsed = None

        self.parse = self.parse_args

        self.string = self.arg = self.add_option
        self.boolean = self.opt = self.true = partial(self.arg, action='store_true')
        self.integer = self.int = partial(self.arg, type=int)
        self.inty = partial(self.arg, action=IntyAction)
        self.strings = partial(self.positional, type=str)

    def __repr__(self):
        return f'<{self.__class__.__name__}>'

    def add_to(self, name=None, group=None, type_=None, default=None, *args, **kwargs):
        args_= args[:]
        kwargs_ = kwargs.copy()
        name_ = name if name else ""
        group_ = group if group else name_
        kwargs_['default'] = default if default else group_ if group_ else name_ if name_ else None
        args_ = [name_[0], name_] + args
        self.add_option(*args_, **kwargs_)

    def add_option(self, initial, name, *args, **kwargs):
        n = name.lstrip('-')
        i = initial.lstrip('-') or n[0]
        return self.parser.add_argument(
            f'-{i}',
            f'--{n}',
            *args, **kwargs)

        if not initial:
            initial = name[0]
        return self.parser.add_argument(
            f'-{initial.lstrip("-")}',
            f'--{name.lstrip("-")}',
            *args, **kwargs)

    @deprecated(reason="use add_option()", version='0.7.33')
    def add_argument(self, initial, name, *args, **kwargs):
        return add_option(initial, name, *args, **kwargs)

    def optional(self, *args, **kwargs):
        """Add an optional positional arg"""
        return self.parser.add_argument(*args, **kwargs, nargs='?')

    def positional(self, *args, **kwargs):
        """Add optional positional args"""
        return self.parser.add_argument(*args, **kwargs, nargs='*')

    def positionals(self, *args, **kwargs):
        """Add mandatory positional args"""
        return self.parser.add_argument(*args, **kwargs, nargs='+')

    def post_parser(self, args):
        """No-op for specialisation"""
        return args

    def parse_args(self, arguments=None, post_parser=None):
        post_parse = post_parser if post_parser else getattr(
            self, 'post_parser', False)
        parsed_args = self.parser.parse_args(arguments)
        argument_namespace = ArgumentsNamespace(parsed_args)
        post_parser_ = post_parse if post_parse else lambda x: x
        posted_args =  post_parser_(argument_namespace)
        self.args = posted_args
        self.args.prog = self.parser.prog
        return self.args

    def parse_string(self, string):
        return self.parse_args(shlex.split(string))


class DescribedParser(ArgumentsParser):
    """An argparse.ArgumentParser with a description, usage and epilog"""
    def __init__(self, description, usage, epilog):
        super().__init__(argparse.ArgumentParser(
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description=description,
            usage=usage,
            epilog=epilog))


class ArgumentsNamespace(object):
    def __init__(self, result):
        self._result = result

    def __getattr__(self, name):
        return getattr(self._result, name)

    def get_args(self):
        attributes = [a for a in dir(self._result) if a[0] != '_']
        return {a: getattr(self._result, a) for a in attributes}

    def get_arg(self, name):
        if not self._result:
            return None
        return getattr(self._result, name)

    def get_strings(self, name):
        if not self._result:
            return None
        return extract_strings(self._result.__dict__, name)

    def set_arg(self, name, value):
        setattr(self._result, name, value)


def parser(description=None, usage=None, epilog=None):
    """Make a command line argument parser"""

    if usage:
        return DescribedParser(description, usage, epilog)
    lines = (description or "").splitlines()
    regexp = re.compile('^[uU]sage:')
    usages = [l for l in lines if regexp.match(l)]
    if usages:
        usage_line = usages.pop()
        usage = usage_line.split(':', 1)[1]
        description = '\n'.join([l for l in lines if l != usage_line])
    elif len(lines) > 1 and not lines[1]:
        usage, description = lines[0], '\n'.join(lines[2:])
    return DescribedParser(description, usage, epilog)



"""Methods for handling lines (of text)"""
import re
from functools import partial


def _chop(lines_in, at, first, last):

    def as_int(string, start, end):
        try:
            return int(string)
        except (ValueError, TypeError):
            lines = lines_in[at:at] if at else lines_in[start:end]
            matcher = re.compile(string)
            for i, line in enumerate(lines, 1 + start):
                if matcher.search(line):
                    return i
        return 0

    def _line(i):
        line = i if i >= 0 else length_read + 1 + i
        if line >= length_read:
            return length_read
        return 0 if line < 1 else line

    def _boundaries():
        if at:
            start = first if first else 0
            end = last if last else -1
            first_ = _line(as_int(at, start, end) - 1)
            return first_, first_ + 1
        first_ = _line(as_int(first, 0, -1) - 1)
        last_ = _line(as_int(last, first_, -1) or -1)
        return first_, last_

    length_read = len(lines_in)
    first, last = _boundaries()
    rest = [] if first > length_read else lines_in[first:last]
    return first, rest


def first_and_rest(text, at=None, first=1, last=-1):
    """First line, and rest of those lines from that text

    at, first, last :
        each can choose a line by number or regexp
        line numbers start at 1, and end at -1
    at: choose this line only
    first: first line to choose
    last: last line to choose
    """
    lines_in = text.splitlines()
    return _chop(lines_in, at, first, last)


def set_width(line, width):
    if not width:
        return line
    return line[:width]

def _number_format(count=999):
    """A string format for line numbers

    Should give a '%d' format with width big enough to `count` lines

    >>> assert _number_format(77) == '%2d: '
    """
    digits = len(str(count))
    return '%%%dd: ' % digits


def add_numbers(lines, first, numbers):
    if not numbers:
        numbered = lambda x, y: x
    else:
        def numbered(line_, line_format_):
            prefix = line_format_ % (first + i + 1)
            return ' '.join((prefix, line_.rstrip()))

    for i, line in enumerate(lines):
        yield numbered(line, _number_format(len(lines)))


def reformat_lines(lines, first, numbers, width):
    return [l[:width] if width else l for l in add_numbers(lines, first, numbers)]


def as_text(lines):
    return '\n'.join(lines)


def select(predicate, lines):
    for line in lines:
        if not predicate(line):
            continue
        yield line

def result(converter, lines):
    for line in lines:
        yield converter(line)

_selector = lambda lambda_, lines_: (i for i in lines_ if lambda_(i))
_converter = lambda lambda_, lines_: (lambda_(i) for i in lines_)

full = lambda lines_: _converter(
    lambda line: line.rstrip(),
    _selector(lambda line: line.rstrip(), lines_)
)
grep = lambda lines_: lambda regexp: _selector(lambda line: line and re.search(regexp, line), lines_)



"""Handle bash commands for pysyte"""

import os
from contextlib import contextmanager
from subprocess import getstatusoutput

from boltons.setutils import IndexedSet


class BashError(ValueError):
    pass


_working_dirs = ['']
_path = []


def cd(path):
    if _working_dirs[0] == path:
        return
    _working_dirs[0] = path


@contextmanager
def pushd(path):
    _working_dirs.insert(0, path)
    yield
    del _working_dirs[0]


def _get_path():
    """Guarantee that /usr/local/bin, /usr/bin, /bin are in PATH"""
    if _path:
        return _path[0]
    environ_paths = IndexedSet(os.environ['PATH'].split(':'))
    minimal_paths = IndexedSet(['/usr/local/bin', '/usr/bin', '/bin'])
    all_paths = environ_paths | minimal_paths
    _path.append(':'.join(all_paths))
    return _path[0]


def run(command):
    path_command = f'PATH={_get_path()} {command}'
    if _working_dirs[0]:
        run_command = f'(cd {_working_dirs[0]}; {path_command})'
    else:
        run_command = path_command
    status, output = getstatusoutput(run_command)
    if status:
        raise BashError(output)
    return output


def full_path(path):
    """Get the real path, expanding links and bashisms"""
    return os.path.realpath(os.path.expanduser(os.path.expandvars(path)))



"""Set up the pysyte project"""

import os
from setuptools import find_packages
from setuptools import setup

import pysyte

description = """Pysyte extends Python

Available on github and pypi for ease of access
"""


def package_files(directory):
    paths = []
    extensions = ('.test', '.tests', '.yaml')
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            extension = os.path.splitext(filename)[-1]
            if extension not in extensions:
                continue
            paths.append(os.path.join('..', path, filename))
    return paths


extra_files = package_files('pysyte')


setup(
    name=pysyte.__name__,
    version=pysyte.__version__,
    description=description.splitlines()[0],
    long_description=description,
    url='https://github.com/jalanb/%s' % pysyte.__name__,
    packages=find_packages(),
    package_data={'': package_files('pysyte')},
    download_url='https://github.com/jalanb/%s/tarball/v%s' % (
        pysyte.__name__, pysyte.__version__),
    license='MIT License',
    author='jalanb',
    author_email='github@al-got-rhythm.net',
    platforms='any',
    classifiers=[
        'Programming Language :: Python :: 3.7',
        'Development Status :: 3 - Alpha',
        'Natural Language :: English',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Build Tools',
    ],
    install_requires=[
        'inflect',
        'pprintpp',
        'path.py==7.7.1',
        'six',
        'stackprinter',
        'pyyaml',
        'boltons',
        'python-magic',
        'deprecated',
    ],
    # entry_points = {
    #     'console_scripts': [
    #         'kat = pysyte.cli.bin:kat'
    #     ]
    # },
    scripts=[
        'bin/kat',
        'bin/getkey',
        'bin/short_dir',
        'bin/imports',
        'bin/rePATH',
        'bin/std',
    ],
    tests_require=['pytest'],
    extras_require={
        'testing': ['pytest'],
    }
)
